import java.awt.*;
import javax.swing.*;
import java.awt.image.BufferedImage;

class HexGrid extends JPanel
{
  private int cells[][];
  private int worldColumns;
  private int worldRows;
  private boolean gridValid;
  private int gridColumns;
  private int gridRows;
  private int xOffset;
  private int yOffset;
  private int cellWidth;
  private int cellHeight;
  private int rowHeight;
  private int gridWidth;
  private int gridHeight;
  private int border;
  private int columnOffset;
  private int rowOffset;
  private int zoomLevel;
  private BufferedImage image;
  private HexagonalCell hexagonalCell;
  private Selector selector;
  private StatusBar statusBar;
  private ColorSelector colorSelector;
  private int currentState;
  private boolean drawSelector;

  public HexGrid(int cells[][], int noColumns, int noRows)
  {
    setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
    this.cells = cells;
    worldColumns = noColumns;
    worldRows = noRows;
    currentState = 1;
    gridValid = false;
    columnOffset = 0;
    rowOffset = 0;
    zoomLevel = 10;
    border = 10;
    drawSelector = false;
  }

  public void setComponents(HexagonalCell hexagonalCell, Selector selector, StatusBar statusBar, ColorSelector colorSelector)
  {
    this.hexagonalCell = hexagonalCell;
    this.selector = selector;    
    this.statusBar = statusBar;
    this.colorSelector = colorSelector;
  }

  public void setCells(int[][] cells)
  {
    this.cells = cells;
  }

  public void makeGrid()
  {
    cellWidth = 2 * zoomLevel;
    cellHeight = 3 * zoomLevel;
    rowHeight = 2 * cellHeight / 3;
    gridWidth = getWidth() - 2 * border;
    gridHeight = getHeight() - 2 * border;
    gridColumns = (gridWidth - cellWidth / 2) / cellWidth;
    if (gridColumns > worldColumns)
      gridColumns = worldColumns;
    xOffset = (getWidth() - (gridColumns * cellWidth + cellWidth / 2)) / 2;
    int h = gridHeight - cellHeight / 3;
    gridRows = h / rowHeight;
    if (gridRows > worldRows)
      gridRows = worldRows;
    yOffset = (getHeight() - (gridRows * rowHeight + cellHeight / 3)) / 2;
    gridValid = true;
    statusBar.setGridSize(gridColumns, gridRows);
  }

  public void setDrawSelector(boolean value)
  {
    drawSelector = value;
  }

  public void paintComponent(Graphics g)
  {
    super.paintComponent(g);
    if ((! gridValid) || (image == null) || (image.getWidth() != getWidth()) || (image.getHeight() != getHeight()))
    {
      image = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_INT_RGB);
      makeGrid();
    }
    statusBar.setGridOffset(columnOffset, rowOffset);
    drawCells();
    if (drawSelector)
      drawSelectedCells();
    g.drawImage(image, 0, 0, null);
    if (! hexagonalCell.isRunning())
    {
      int p = 0;
       for (int c = 0; c < worldColumns; c++)
         for (int r = 0; r < worldRows; r++)
          if ((cells[c][r] & selector.stateMask) > 0)
            p++;
      statusBar.setPopulation(p);
    }
  }

  private Polygon getPolygon(int x, int y)
  {
    int x1 = x + cellWidth / 2;
    int x2 = x + cellWidth;
    int y1 = y + cellHeight / 3;
    int y2 = y + 2 * cellHeight /3;
    int y3 = y + cellHeight;
    Polygon polygon = new Polygon();
    polygon.addPoint(x, y1);
    polygon.addPoint(x1, y);
    polygon.addPoint(x2, y1);
    polygon.addPoint(x2, y2);
    polygon.addPoint(x1, y3);
    polygon.addPoint(x, y2);
    return polygon;
  }

  public int offsetColumn(int c)
  {
    return (c + columnOffset) % worldColumns;
  }

  public int offsetRow(int r)
  {
    return (r + rowOffset) % worldRows;
  }

  public int worldColumnToGridColumn(int worldColumn)
  {
    return (worldColumns + worldColumn - columnOffset) % worldColumns;
  }

  public int worldRowToGridRow(int worldRow)
  {
    return (worldRows + worldRow - rowOffset) % worldRows;
  }

  private void drawCells()
  {
    Graphics g = image.getGraphics();
    g.clearRect(0, 0, image.getWidth(), image.getHeight());
    for (int i = 0; i < gridColumns; i++)
      for (int j = 0; j < gridRows; j++)
      {
        int xo = xOffset;
        if (j % 2 > 0)
          xo += cellWidth / 2;
        Polygon p = getPolygon(xo + i * cellWidth, yOffset + j * rowHeight);
        g.setColor(colorSelector.getColor(cells[offsetColumn(i)][offsetRow(j)] & selector.stateMask));
        g.fillPolygon(p);
        if (zoomLevel > 2)
        { 
          g.setColor(colorSelector.getGridColor());
          g.drawPolygon(p);
        }
      }
    g.dispose();
  }

  private void drawSelectedCells()
  {
    Graphics g = image.getGraphics();
    for (int i = 0; i < gridColumns; i++)
      for (int j = 0; j < gridRows; j++)
      {
        int xo = xOffset;
        if (j % 2 > 0)
          xo += cellWidth / 2;
        if ((cells[offsetColumn(i)][offsetRow(j)] & selector.selected) != 0)
        {
          Polygon p = getPolygon(xo + i * cellWidth, yOffset + j * rowHeight);
          int c = selector.getCell(i, j);
          if (c != 0)
          {
            g.setColor(colorSelector.getColor(c));
            g.fillPolygon(p);
          }
          g.setColor(colorSelector.getSelectorColor());
          g.drawPolygon(p);
        }
      }
    g.dispose();
  }

  public void changeCurrentState()
  {
    currentState = (currentState + 1) % hexagonalCell.getNoStates();
  }

  public void setPosition(int x, int y)
  {
    for (int i = 0; i < gridColumns; i++)
      for (int j = 0; j < gridRows; j++)
      {
        int xo = xOffset;
        if (j % 2 > 0)
          xo += cellWidth / 2;
        if (getPolygon(xo + i * cellWidth, yOffset + j * rowHeight).contains(x, y))
        {
          statusBar.setPosition(offsetColumn(i), offsetRow(j), i, j);
          statusBar.setState(cells[offsetColumn(i)][offsetRow(j)] & selector.stateMask);
          statusBar.repaint();
          return;
        }
      }
    statusBar.removePosition();
    statusBar.removeState();
    statusBar.repaint();
  }

  public void setCellState(int x, int y)
  {
    for (int i = 0; i < gridColumns; i++)
      for (int j = 0; j < gridRows; j++)
      {
        int xo = xOffset;
        if (j % 2 > 0)
          xo += cellWidth / 2;
        if (getPolygon(xo + i * cellWidth, yOffset + j * rowHeight).contains(x, y))
        {
          cells[offsetColumn(i)][offsetRow(j)] = currentState;
          statusBar.setState(cells[offsetColumn(i)][offsetRow(j)] & selector.stateMask);
          repaint();
          return;
        }
      }
  }

  public Point selectCell(int x, int y)
  {
    for (int i = 0; i < gridColumns; i++)
      for (int j = 0; j < gridRows; j++)
      {
        int xo = xOffset;
        if (j % 2 > 0)
          xo += cellWidth / 2;
        if (getPolygon(xo + i * cellWidth, yOffset + j * rowHeight).contains(x, y))
          return new Point(i, j);
      }
    return null;
  }

  public void incColumnOffset(int inc)
  {
    columnOffset = (columnOffset + inc) % worldColumns;
    if (selector.getState() != Selector.State.None)
      selector.decTopLeftColumn(inc);
    repaint();
  }

  public void decColumnOffset(int inc)
  {
    columnOffset = (worldColumns + columnOffset - inc) % worldColumns;
    if (selector.getState() != Selector.State.None)
      selector.incTopLeftColumn(inc);
    repaint();
  }

  public void incRowOffset(int inc)
  {
    rowOffset = (rowOffset + inc) % worldRows;
    if (selector.getState() != Selector.State.None)
      selector.decTopLeftRow(inc);
    repaint();
  }

  public void decRowOffset(int inc)
  {
    rowOffset = (worldRows + rowOffset - inc) % worldRows;
    if (selector.getState() != Selector.State.None)
      selector.incTopLeftRow(inc);
    repaint();
  }

  public void resetOffset()
  {
    columnOffset = 0;
    rowOffset = 0;
  }

  public void zoomIn()
  {
    zoomLevel++;
    gridValid = false;
    repaint();
  }

  public void zoomOut()
  {
    if (zoomLevel > 1)
    {
      zoomLevel--;
      gridValid = false;
      repaint();
    }
  }

  public void setCurrentState(int state)
  {
    currentState = state;
  }

  public int getCurrentState()
  {
    return currentState;
  }

  public int getGridColumns()
  {
    return gridColumns;
  }

  public int getGridRows()
  {
    return gridRows;
  }
}